@@ -19,7 +19,7 @@
 # TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 # SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-"""\ 
+"""\
 EasyCluster is a remote execution / clustering module for Python. 
  
 To get started, first generate a keyfile: 
@@ -74,7 +74,7 @@ from types import GeneratorType, ModuleType, FunctionType, BuiltinFunctionType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.16'
+VERSION = '0.17'
 MIN_COMPAT_VERSION = '0.16'
 
 PYTHON3 = sys.version_info >= (3, 0)
@@ -86,18 +86,26 @@ if PYTHON3:
         return b2a_hex(data).decode('ascii')
     def _hexdec(data):
         return a2b_hex(data.encode('ascii'))
+    def _str_to_bytes(data):
+        return str(data).encode('utf-8')
+
     _next_method = '__next__'
+
+    DEFAULT_PORT = 11998
 else:
     import cPickle as pickle
     import Queue as queue
     import __builtin__ as builtins
     _hexenc = b2a_hex
     _hexdec = a2b_hex
+    _str_to_bytes = str
 
     ## Make all classes new-style by default
     __metaclass__ = type
     _next_method = 'next'
 
+    DEFAULT_PORT = 11999
+
 try:
     from select import epoll
 except ImportError:
@@ -131,8 +139,6 @@ HMAC_SIZE = len(HMAC_MOD().digest())
 ERROR, CHALLENGE, CHALLENGE_RESPONSE, KEEPALIVE, REQUEST, RESPONSE, INIT = range(7)
 _pickle_messages = set([REQUEST, RESPONSE, INIT])
 
-DEFAULT_PORT = 11999
-
 _singletons = {}
 
 _extra_base_classes = {}
@@ -240,6 +246,7 @@ def make_singleton(obj, export_methods=('@auto',), export_attrs=('@auto',),
     '''
 
     addr = id(obj)
+    typ = obj.__class__
     def wrcb(wr, addr=addr):
         _singletons.pop(addr, None)
         
@@ -249,9 +256,9 @@ def make_singleton(obj, export_methods=('@auto',), export_attrs=('@auto',),
     except TypeError:
         ## type(obj) is not weakref-capable.
         wr = None
-        
+
     typedat = _make_typedat(
-        obj.__class__, export_methods, export_attrs, export_attrs_cache, proxy_class)
+        typ, export_methods, export_attrs, export_attrs_cache, proxy_class)
     _singletons[addr] = typedat, wr, obj
 
 class DummyConnection(object):
@@ -385,7 +392,8 @@ class RemoteProxy(RemoteProxyBase):
         See Client.send_request for the meaning of the 'nonblocking',
         'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
-        return self._connection.send_request(0, 'getattr', (self, name))
+        return self._connection.send_request(0, 'getattr', (self, attr),
+                                             nonblocking, oncomplete, onerror, threadid, origexc)
 
     def raw_set_attribute(self, attr, val, kwargs={}, nonblocking=False,
                         oncomplete=None, onerror=None, threadid=None,
@@ -399,7 +407,8 @@ class RemoteProxy(RemoteProxyBase):
         See Client.send_request for the meaning of the 'nonblocking',
         'oncomplete', 'onerror', 'threadid', and 'origexc' arguments. 
         '''
-        return self._connection.send_request(0, 'setattr', (self, name, val))
+        return self._connection.send_request(0, 'setattr', (self, attr, val),
+                                             nonblocking, oncomplete, onerror, threadid, origexc)
 
     def __repr__(self):
         return '<RemoteProxy for oid %d on %s>' % (self._oid, self._connection.peername)
@@ -727,7 +736,7 @@ class Connection:
                         try:
                             return self._got_init(pick)
                         except ValueError as e:
-                            self.send_message(ERROR, bytes(str(e)))
+                            self.send_message(ERROR, _str_to_bytes(e))
                             raise
 
                 elif rmsg == KEEPALIVE:
@@ -1266,8 +1275,8 @@ class _ATCThread(threading.Thread):
                     if atc._read_data(False):
                         ## Check thread status after every valid message received
                         atc.check_threads()
-        except (socket.error, select.error) as e:
-            if e[0] != errno.ECONNRESET and e[0] != errno.EBADF:
+        except (EnvironmentError, select.error) as e:
+            if e.args[0] != errno.ECONNRESET and e.args[0] != errno.EBADF:
                 raise
         except EOFError:
             ## Swallow this error so it isn't printed. Since we call close(),
@@ -1502,7 +1511,7 @@ def decode_key(key):
             return _hexdec(key)
     except Exception:
         pass
-    return HMAC_MOD(key.encode('utf-8')).digest()
+    return HMAC_MOD(_str_to_bytes(key)).digest()
 
 def read_key_file(path):
     '''Read an HMAC key from a file.'''
@@ -1735,7 +1744,10 @@ def _make_typedat(typ, export_methods, export_attrs, export_attrs_cache, proxy_c
                 if cls is not object:
                     attrs.update(cls.__dict__)
 
-            functypes = (FunctionType, BuiltinFunctionType, staticmethod, classmethod)
+            # Built-in slot methods have a type of <type 'wrapper_descriptor'>
+            # that does not have a name exported in the types module.
+            functypes = (FunctionType, BuiltinFunctionType, staticmethod, 
+                         classmethod, type(FunctionType.__dict__['__call__']))
             export_methods = [name for name, val in attrs.items()
                               if (name not in _forbidden_slots and 
                                   isinstance(val, functypes))]
@@ -1848,10 +1860,10 @@ if sys.platform.startswith('win'):
                                        win32con.CREATE_NEW, 0, 0)
         except pywintypes.error as e:
             ## File exists.
-            if e[0] == 80:
+            if e.args[0] == 80:
                 return False
         try:
-            win32file.WriteFile(fil, _hexenc(key) + '\r\n', None)
+            win32file.WriteFile(fil, _str_to_bytes(_hexenc(key) + '\r\n'), None)
         finally:
             win32file.CloseHandle(fil)
         return True
@@ -1872,12 +1884,12 @@ else:
         try:
             fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o600)
         except OSError as e:
-            if e[0] == errno.EEXIST:
+            if e.args[0] == errno.EEXIST:
                 return False
             raise
 
         try:
-            os.write(fd, _hexenc(key) + '\n')
+            os.write(fd, _str_to_bytes(_hexenc(key) + '\n'))
         finally:
             os.close(fd)
         return True
