@@ -149,7 +149,7 @@ def _set_global(name, defmod, value):
         gv = _get_global('.'.join(components[:-1]), defmod)
         setattr(gv, components[-1], value)
 
-_method_types = BuiltinMethodType, MethodType, FunctionType
+_method_types = BuiltinMethodType, MethodType, FunctionType, type(_set_global.__call__)
 
 class RootObject(object):
     '''Utility class instantiated automatically on the server as OID 0.'''
@@ -390,6 +390,79 @@ class QuietServer(Server):
     def report_request(self, *args):
         pass
 
+class IPAddress(object):
+    '''Represents an IP address with optional network mask length'''
+    def __init__(self, txt):
+        maskbits = None
+        lst = txt.split('/', 1)
+        addr = lst[0]
+        if len(lst) == 2:
+            maskbits = int(lst[1])
+        
+        if ':' in addr:
+            components = addr.split(':')
+
+            if '.' in components[-1]:
+                v4part = IPAddress(components.pop()).aslong
+                components.append('%04x' % (v4part >> 16))
+                components.append('%04x' % (v4part & 0xFFFF))
+
+            if not 2 <= len(components) <= 8:
+                raise ValueError('Invalid IPv6 address %r' % addr)
+
+            val = 0
+            curword = 0
+            for i, part in enumerate(components):
+                if part == '':
+                    if i == 0:
+                        curword += 1
+                    else:
+                        curword += 9 - len(components)
+                else:
+                    cval = int(part, 16)
+                    if not 0 <= cval <= 65535:
+                        raise ValueError('Invalid component %r in IPv6 address %r' % (part, addr))
+                    val |= cval << (16 * (7 - curword))
+                    curword += 1
+
+            if val & ~0xFFFFFFFF == 0xFFFF00000000:
+                val &= 0xFFFFFFFF
+                self.family = socket.AF_INET
+                numbits = 32
+            else:
+                self.family = socket.AF_INET6
+                numbits = 128
+            self.aslong = val
+        else:
+            components = addr.split('.')
+            if len(components) != 4:
+                raise ValueError('Invalid IPv4 address: %r' % addr)
+
+            self.family = socket.AF_INET
+            shift = 24
+            val = 0
+            for part in components:
+                cval = int(part)
+                if not 0 <= cval <= 255:
+                    raise ValueError('Invalid component %r in IPv4 address %r' % (part, addr))
+                val |= cval << shift
+                shift -= 8
+            self.aslong = sum(int(v) << (24 - i * 8) for i, v in enumerate(components))
+            numbits = 32
+
+        if maskbits is None:
+            maskbits = numbits
+        elif not 0 <= maskbits <= numbits:
+            raise ValueError('Invalid value for mask length %r in address %r: '
+                             'must be between 0 and %d' % (maskbits, txt, numbits))
+        self.maskbits = maskbits
+        self.mask = (1 << numbits) - (1 << (numbits - maskbits))
+        self.aslong &= self.mask
+
+    def in_network(self, net):
+        return (self.family == net.family and 
+                (self.aslong & net.mask) == net.aslong)
+
 def spawn_local(clientclass=_core.Client, svrclass=QuietServer):
     '''Create a private server instance and connect to it.'''
     global _local_id
@@ -430,6 +503,9 @@ def server_main(args=None):
     options.add_option('-O', '--overwrite', action='store_true', help='Allow --generate to overwrite existing keyfile')
     options.add_option('-p', '--port', type='int', default=_core.DEFAULT_PORT, help='TCP port to listen on. Default: %d' % _core.DEFAULT_PORT)
     options.add_option('-c', '--class', dest='svrclass', default='Server', help='Fully-qualified name of server class')
+    options.add_option('-b', '--bind', default=None, help='Bind the server to a specific IP address')
+    options.add_option('-a', '--allow', action='append', default=[], help='Allow only a specific address or subnet to connect '
+                       '(e.g. 192.168.0.0/24). Specify more than once to allow multiple addresses.')
 
     if _daemonize:
         options.add_option('-d', '--daemonize', action='store_true',
@@ -498,7 +574,7 @@ def server_main(args=None):
 
     elif opts.serve:
         key = _core.key_from_options(opts)
-        if key == b'' and warn:
+        if key == b'':
             print('SECURITY WARNING: running with blank key.', file=sys.stderr)
         if '.' in opts.svrclass:
             svrmod, svrclassname = opts.svrclass.rsplit('.')
@@ -506,13 +582,14 @@ def server_main(args=None):
         else:
             clas = globals()[opts.svrclass]
 
-        run_server(opts.port, key, clas, 
+        run_server(opts.port, key, opts.bind, opts.allow, clas,
                    getattr(opts, 'daemonize', False),
                    getattr(opts, 'pidfile', None))
     else:
         options.print_help()
 
-def run_server(port, key, svrclass=Server, daemonize=False, pidfile=None):
+def run_server(port, key, bindaddr=None, allow=(), 
+               svrclass=Server, daemonize=False, pidfile=None):
     '''Runs a server on the specified port.
 
     port is the TCP port to listen on.
@@ -522,23 +599,35 @@ def run_server(port, key, svrclass=Server, daemonize=False, pidfile=None):
     '''
     global _server_socket
 
+    allow = [IPAddress(v) for v in allow]
+
     server_init()
 
     if socket.has_ipv6:
-        sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
+        gai_family = 0
+        family = socket.AF_INET6
+    else:
+        gai_family = family = socket.AF_INET
+
+    sockaddr = ('', port)
+    if bindaddr is not None:
+        addrs = socket.getaddrinfo(bindaddr, port, gai_family)
+        family, socktype, proto, canonname, sockaddr = addrs[0]
+
+    sock = socket.socket(family, socket.SOCK_STREAM)
+
+    if family == socket.AF_INET6 and socket.has_ipv6:
         # Windows needs this socket option set to listen on both IPV4 and
         # IPV6. However, IPPROTO_IPV6 is missing from the socket module on Windows.
         # Since the value is likely to never change, just include it here.
         IPPROTO_IPV6 = 41
         sock.setsockopt(IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
-    else:
-        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  
     sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
     
     _core._mark_non_inheritable(sock.fileno())
 
-    sock.bind(('', port))
+    sock.bind(sockaddr)
     sock.listen(1)
     _server_socket = sock
 
@@ -550,21 +639,34 @@ def run_server(port, key, svrclass=Server, daemonize=False, pidfile=None):
     while _server_socket:
         try:
             r, w, e = select.select([sock], (), (), 2)
-        except (socket.error, select.error) as e:
-            if e[0] not in _core._nonblocking_errors:
+        except (EnvironmentError, select.error) as e:
+            if e.args[0] not in _core._nonblocking_errors:
                 raise
             
         if not r:
             continue
         try:
             csock, addr = sock.accept()
-        except socket.error as e:
-            if e.errno in _core._nonblocking_errors:
+        except EnvironmentError as e:
+            if e.args[0] in _core._nonblocking_errors:
                 continue
             else:
                 raise
-        spawn_runner_process(csock, _core._get_peer_name(addr), svrclass, key)
-        csock.close()
+        try:
+            if allow:
+                ipaddr = IPAddress(addr[0])
+                for net in allow:
+                    if ipaddr.in_network(net):
+                        break
+                else:
+                    print('Connection from %s not allowed' % addr[0])
+                    continue
+
+            spawn_runner_process(csock, _core._get_peer_name(addr), svrclass, key)
+        except Exception:
+            traceback.print_exc()
+        finally:
+            csock.close()
         
             
 def stop_server():
